/*

------------------------------------------------------------------------------

A license is hereby granted to reproduce this software source code and
to create executable versions from this source code for personal,
non-commercial use.  The copyright notice included with the software
must be maintained in all copies produced.

THIS PROGRAM IS PROVIDED "AS IS". THE AUTHOR PROVIDES NO WARRANTIES
WHATSOEVER, EXPRESSED OR IMPLIED, INCLUDING WARRANTIES OF
MERCHANTABILITY, TITLE, OR FITNESS FOR ANY PARTICULAR PURPOSE.  THE
AUTHOR DOES NOT WARRANT THAT USE OF THIS PROGRAM DOES NOT INFRINGE THE
INTELLECTUAL PROPERTY RIGHTS OF ANY THIRD PARTY IN ANY COUNTRY.

Copyright (c) 1994-2006, John Conover, All Rights Reserved.

Comments and/or bug reports should be addressed to:

    john@email.johncon.com (John Conover)

    http://www.johncon.com/ntropix/
    http://www.johncon.com/

------------------------------------------------------------------------------

spcid.c is for using a modem to display telephone line Caller
Identification, (CID,) data and dial telephone numbers.  The program
uses synchronous canonical serial port processing.  The default port
is /dev/ttyS0 at 38400 baud. The default semaphore lock file is
/var/lock/LCK..ttyS0, which has an fcntl(2) advisory lock applied,
also. The serial port is controlled through termios(3), and is
restored to original values on exit. The serial port child read
process is created by fork(2) from the parent write process, and both
processes respond to SIGINT = ^C and SIGTERM signals.

The CID data from the modem is formatted:

    DATE = 1102
    TIME = 2036
    NAME = Someone's Name
    NMBR = 1235551212

The CID data is formatted and printed to stdout, (or possibly a named
pipe specified with the outfile argument,) and formatted and appended
to a log file, (with the default name /var/log/cid/spcid.log,) as a
Unix tab delimited flat file database. There is a simple Perl(1)
script, spcid.pl, to query the database.

Phone numbers can be typed on the program's stdin, (or possibly a
named pipe specified with the infile argument,) for the modem to place
dial calls. The modem dials the number, and immediately goes on hook,
so a telephone on the same line must be off hook by the end of the
dial-which is useful for placing calls from a database, or computer
terminal. (The proceedure is to lift the receiver off hook, and then
have the program dial the number, possibly out of a menu, etc.)

$Revision: 0.0 $
$Date: 2006/06/18 19:36:00 $
$Id: spcid.c,v 0.0 2006/06/18 19:36:00 john Exp $
$Log: spcid.c,v $
Revision 0.0  2006/06/18 19:36:00  john
Initial revision


*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <termios.h>
#include <signal.h>
#include <ctype.h>
#include <time.h>
#include <sys/stat.h>
#include <sys/wait.h>

#define _POSIX_SOURCE 1 /* POSIX compliant source */

#define AREACODE "000" /* local area code, will be stripped from dialed phone numbers */
#define BAUDRATE B38400 /* baud rate for serial port communication */
#define SPDEVICE "/dev/ttyS0" /* default serial port device */
#define LOGFILE "/var/log/cid/spcid.log" /* default log file name */
#define LOCKFILE "/var/lock/LCK..ttyS0" /* default serial port lock file name */
#define SPSTRING "AT&F1E1V1Q0S0=0#CID=1" /* default serial port string for CID */
#define SPEOL "\r" /* serial port end-of-line string */
#define DIALPREFIX "ATDT" /* modem dial prefix to dial a number */
#define DIALSUFFIX ";H0" /* modem dial suffix to return to command mode after dialing a number */

#define BUFLEN BUFSIZ /* size of record buffers */
#define TOKEN_SEPARATORS " \f\n\r\t\v\b" /* file record field separators, which is white space */
#define NUMBER_TOKEN_SEPARATORS " \f\n\r\t\v\b()-+." /* phone number field separators, which is white space, () and -, as in "+1 (321) 987-6543", or "321.987.6543" */

static char rcsid[] = "$Id: spcid.c,v 0.0 2006/06/18 19:36:00 john Exp $"; /* program version */
static char copyright[] = "Copyright (c) 1994-2006, John Conover, All Rights Reserved"; /* the copyright banner */

#ifdef __STDC__

static const char *help_message[] = /* help message index array */

#else

static char *help_message[] = /* help message index array */

#endif

{
    "\n",
    "Use a modem to display telephone line Caller ID\'s and dial telephone numbers\n",
    "Usage: spcid [-D] [-d device] [-i infilename] [-L logfile] [-l lockfile]\n             [-o outfilename] [-s string] [-v]\n",
    "    -D, enable debugging, (disabled)\n",
    "    -d device, device name, (/dev/ttyS0)\n",
    "    -i infilename, input filename, (stdin)\n",
    "    -L logfile, log file name, (/var/log/cid/cid.log)\n",
    "    -l lockfile, lock file name, (/var/lock/LCK..ttyS0)\n",
    "    -o outfilename, output file name, (stdout)\n",
    "    -s string, serial port initialization string, (AT&F1E1V1Q0S0=0#CID=1)\n",
    "    -v, print the program's version information\n"
};

#ifdef __STDC__

static const char *error_message[] = /* error message index array */

#else

static char *error_message[] = /* error message index array */

#endif

{
    "No error\n",
    "Error in program argument(s)\n",
    "Error opening serial device\n",
    "Error getting serial device attributes\n",
    "Error installing signal handler\n",
    "Error flushing serial device attributes\n",
    "Error setting serial device attributes\n",
    "Error writing to serial device\n",
    "Error resetting serial device attributes\n",
    "Error closing serial device\n",
    "Error opening input file\n",
    "Error closing input file\n",
    "Error creating lock file\n",
    "Error removing lock file\n",
    "Error forking to child read process\n",
    "Error reading from serial device\n",
    "Error locking lock file\n",
    "Error writing to lock file\n",
    "Error opening log file\n",
    "Error writing to log file\n",
    "Error closing log file\n",
    "Error processing dial record\n",
    "Error getting Epoc time\n",
    "Error opening output file\n",
    "Error closing output file\n",
    "Error write buffer overflow\n",
    "Error reading input file\n",
    "Error read buffer overflow\n",
    "Error writing to output file\n",
    "Error enabling blocking on serial device\n",
    "Error flushing output file\n",
    "Error setting baud rate\n",
    "Error flushing serial device\n"
};

#define NOERROR 0 /* negative error values, one for each index in the error message array */
#define EARGS -1
#define EOPEN -2
#define EGATTR -3
#define ESIG -4
#define EFLUSH -5
#define ESATTR -6
#define EWRITE -7
#define ERATTR -8
#define ECLOSE -9
#define EIOPEN -10
#define EICLOSE -11
#define ELOCKF -12
#define ELOCKR -13
#define EFORK -14
#define EREAD -15
#define EFCNL -16
#define EWLCK -17
#define ELOGO -18
#define ELOGW -19
#define ELOGC -20
#define EPDIAL -21
#define EEPOC -22
#define EOOPEN -23
#define EOCLOSE -24
#define EWROVFL -25
#define ERDIN -26
#define EREOVFL -27
#define EOWRITE -28
#define EBLOCK -29
#define EOFLUSH -30
#define EBAUD -31
#define EDFLUSH -32

#ifdef __STDC__

static void print_message (int pretval); /* print any error messages */
static void spclose (int err); /* interrupt handler */
static int child_read (int fdd); /* child read loop, process all read records from the modem */
static int parent_write (int fdd, int infiled); /* parent write loop, process all write records to the modem */
static int strtoken (char *string, char *parse_array, char **parse, const char *delim); /* parse a record based on sequential delimiters */

#else

static void print_message (); /* print any error messages */
static void spclose (); /* interrupt handler */
static int child_read (); /* child read loop, process all read records from the modem */
static int parent_write (); /* parent write loop, process all write records to the modem */
static int strtoken (); /* parse a record based on sequential delimiters */

#endif

static char serial_logname[] = {LOGFILE}, /* default log file name */
            *logname = serial_logname, /* default log file name */
            serial_lockname[] = {LOCKFILE}, /* default lock file name */
            *lockname = serial_lockname, /* default lock file name */
            serial_device[] = {SPDEVICE}, /* default serial port device name */
            serial_string[] = {SPSTRING}, /* default serial port dialer string */
            dialprefix[] = {DIALPREFIX}, /* modem dial prefix to dial a number */
            dialsuffix[] = {DIALSUFFIX}, /* modem dial suffix to return to command mode after dialing a number */
            eol[] = {SPEOL}; /* serial port end-of-line string */

static int retval = NOERROR, /* return value, assume no error */
           fd = -1, /* serial port device file descriptor */
           infile = STDIN_FILENO, /* input file descriptor */
           semaphorelock = -1, /* lock file file descriptor */
           state = 0, /* state of serial port setup, 0 = not set up, 1 = original device attributes successfully saved */
           debug = 0, /* debug flag, 0 = no, 1 = yes */
           remove_lockfile = 0; /* remove lock file, 0 = no, 1 = yes */

static FILE *logfile = (FILE *) 0, /* reference to log file */
            *outfile = (FILE *) 0; /* reference to output file */

static pid_t pid_write = (pid_t) 0, /* pid of parent write process */
             pid_read = (pid_t) 0; /* pid of child read process */

static struct termios oldtio; /* original device attributes, restored on exit */

static struct flock lockfile; /* lock file structure */

#ifdef __STDC__

int main (int argc, char *argv[])

#else

int main (argc, argv)
int argc;
char *argv[];

#endif

{
    static char pid_str[BUFLEN]; /* pid string */

    char *device = serial_device, /* default serial port device name */
         *string = serial_string, /* default serial port dialer string */
         *in_name = (char *) 0, /* input file name */
         *out_name = (char *) 0; /* output file name */

    int c; /* command line switch */

    struct termios newtio; /* device attributes used by this program */

    pid_write = getpid (); /* save the pid of parent write process */

    outfile = (FILE *) stdout; /* reference to output file, (can not be initialized as a global) */

    while ((c = getopt (argc, argv, "Dd:i:L:l:o:s:v")) != EOF) /* for each command line switch */
    {

        switch (c) /* which switch? */
        {

            case 'D': /* request for debugging? */

                debug = 1; /* debug flag, 0 = no, 1 = yes */
                break;

            case 'd': /* request for device name? */

                device = optarg; /* yes, save the device name */
                break;

            case 'i': /* request for input file name? */

                in_name = optarg; /* yes, save the input file name */
                break;

            case 'L': /* request for log file name? */

                logname = optarg; /* yes, save the log file name */
                break;

            case 'l': /* request for lock file name? */

                lockname = optarg; /* yes, save the lock file name */
                break;

            case 'o': /* request for output file name? */

                out_name = optarg; /* yes, save the output file name */
                break;

            case 's': /* request for device dialer string? */

                string = optarg; /* yes, save the device dialer string */
                break;

            case 'v':

                (void) printf ("%s\n", rcsid); /* print the version */
                (void) printf ("%s\n", copyright); /* print the copyright */
                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */

            default: /* illegal switch? */

                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */
                break;
        }

    }

    if (retval == NOERROR) /* enough arguments? */
    {
        retval = ESIG; /* assume error installing signal handler */

        if (signal (SIGINT, spclose) != SIG_ERR) /* install the interrupt handler for SIGINT = ^C */
        {

            if (signal (SIGTERM, spclose) != SIG_ERR) /* install the interrupt handler for SIGTERM */
            {

                if (signal (SIGCHLD, spclose) != SIG_ERR) /* install the interrupt handler for SIGCHLD */
                {
                    lockfile.l_type = F_WRLCK; /* exclusive write lock the entire file */
                    lockfile.l_whence = (short int) SEEK_SET;
                    lockfile.l_start = (off_t) 0;
                    lockfile.l_len = (off_t) 0;

                    retval = ELOCKF; /* assume error creating lock file */

                    if ((semaphorelock = open (lockname, O_RDWR | O_CREAT | O_EXCL, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH)) >= 0) /* open the lock file exclusively */
                    {
                        remove_lockfile = 1; /* set the remove lock file, 0 = no, 1 = yes */
                        retval = NOERROR; /* assume no error */
                    }

                    else if ((semaphorelock = open (lockname, O_RDWR, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH)) >= 0) /* open the lock file if it exists */
                    {
                        retval = NOERROR; /* assume no error */
                    }

                    if (retval == NOERROR) /* any errors? */
                    {
                        retval = EFCNL; /* assume error locking lock file */

                        if (fcntl (semaphorelock, F_SETLK, &lockfile) == 0) /* exclusive write lock the lock file */
                        {
                            remove_lockfile = 1; /* set the remove lock file, 0 = no, 1 = yes */
                            retval = EWLCK; /* assume error writing to lock file */

                            if (ftruncate (semaphorelock, 0) == 0) /* truncate the lock file */
                            {
                                (void) sprintf (pid_str, "%10d\n", pid_write); /* the pid string is the PID right justified to 10 characters plus a '\n' EOL */

                                if (write (semaphorelock, pid_str, strlen (pid_str)) >= 0) /* write the serial port dialer string to the serial port */
                                {
                                    retval = EOPEN; /* assume error opening device */

                                    if ((fd = open (device, O_RDWR | O_NOCTTY | O_NDELAY)) >= 0) /* open the serial port device file descriptor */
                                    {
                                        retval = EBLOCK; /* assume error enabling blocking on serial device */

                                        if (fcntl (fd, F_SETFL, 0) == 0) /* enable read blocking, i.e., read () waits until a character is in the buffer */
                                        {
                                            retval = EFLUSH; /* assume error flushing device attributes */

                                            if (tcflush (fd, TCIOFLUSH) == 0) /* flush device input/output */
                                            {
                                                retval = EGATTR; /* assume error getting device attributes */

                                                if (tcgetattr (fd, &oldtio) == 0) /* save original port settings */
                                                {
                                                    state = 1; /* state of serial port setup, 0 = not set up, 1 = original device attributes successfully saved */
                                                    memcpy (&newtio, &oldtio, sizeof (newtio)); /* copy the original device attributes */
                                                    retval = EBAUD; /* assume error setting baud rate */

                                                    if (cfsetispeed (&newtio, BAUDRATE) == 0) /* set the baud rate */
                                                    {

                                                        if (cfsetospeed (&newtio, BAUDRATE) == 0)
                                                        {

                                                            /* terminal set up */

                                                            newtio.c_cflag |= CRTSCTS; /* HW flow control */
                                                            newtio.c_lflag |= (ICANON | ISIG); /* canonical input */
                                                            newtio.c_lflag &= ~(ECHO | ECHOE); /* no local echo */
                                                            newtio.c_oflag |= OPOST; /* canonical output */
                                                            newtio.c_iflag |= IGNCR;

                                                            newtio.c_cflag |= (CLOCAL | CREAD); /* local mode, enable receiver */
                                                            newtio.c_cflag &= ~CSIZE; /* Mask the character size bits */
                                                            newtio.c_cflag |= CS8; /* Select 8 data bits */
                                                            newtio.c_cflag &= ~PARENB; /* No parity (8N1) */
                                                            newtio.c_cflag &= ~CSTOPB;
                                                            newtio.c_cc[VINTR] = 0; /* Ctrl-c */
                                                            newtio.c_cc[VQUIT] = 0; /* Ctrl-\ */
                                                            newtio.c_cc[VERASE] = 0; /* del */
                                                            newtio.c_cc[VKILL] = 0; /* @ */
                                                            newtio.c_cc[VEOF] = 4; /* Ctrl-d */
                                                            newtio.c_cc[VTIME] = 0; /* inter-character timer unused */
                                                            newtio.c_cc[VMIN] = 1; /* blocking read until 1 character arrives */
                                                            newtio.c_cc[VSWTC] = 0; /* '\0' */
                                                            newtio.c_cc[VSTART] = 0; /* Ctrl-q */
                                                            newtio.c_cc[VSTOP] = 0; /* Ctrl-s */
                                                            newtio.c_cc[VSUSP] = 0; /* Ctrl-z */
                                                            newtio.c_cc[VEOL] = 0; /* '\0' */
                                                            newtio.c_cc[VREPRINT] = 0; /* Ctrl-r */
                                                            newtio.c_cc[VDISCARD] = 0; /* Ctrl-u */
                                                            newtio.c_cc[VWERASE] = 0; /* Ctrl-w */
                                                            newtio.c_cc[VLNEXT] = 0; /* Ctrl-v */
                                                            newtio.c_cc[VEOL2] = 0; /* '\0' */

                                                            retval = ESATTR; /* assume error setting device attributes */

                                                            if (tcsetattr (fd, TCSANOW, &newtio) == 0) /* set the device attributes */
                                                            {
                                                                retval = EFLUSH; /* assume error flushing device attributes */

                                                                if (tcflush (fd, TCIOFLUSH) == 0) /* flush device input/output */
                                                                {
                                                                    retval = EWRITE; /* assume error writing to device */

                                                                    if (write (fd, string, strlen (string)) >= 0) /* write the serial port dialer string to the serial port */
                                                                    {

                                                                        if (write (fd, eol, strlen (eol)) >= 0) /* write the CR of serial port dialer string to the serial port */
                                                                        {
                                                                            retval = EOOPEN; /* assume error opening output file */

                                                                            if ((outfile = (out_name == (char *) 0) ? outfile : fopen (out_name, "w")) != (FILE *) 0) /* open the output file */
                                                                            {
                                                                                retval = EIOPEN; /* assume error opening file */

                                                                                if ((infile = (in_name == (char *) 0) ? infile : open (in_name, O_RDWR | O_NOCTTY)) >= 0) /* open the input file */
                                                                                {
                                                                                    retval = NOERROR; /* assume no error */

                                                                                    if ((pid_read = fork ()) < 0) /* fork the child read process */
                                                                                    {
                                                                                        retval = EFORK; /* assume error forking to child read process */
                                                                                    }

                                                                                    else if (pid_read == 0) /* child or parent process, (pid_read is zero in the child process, the child process' PID in the parent)? */
                                                                                    {

                                                                                        /* start of child read process; pid_write still contains the PID of the parent write process */

                                                                                        retval = child_read (fd); /* child read loop, process all read records from the modem */

                                                                                        /* end of the child read process, fall through and exit */
                                                                                    }

                                                                                    else /* wasn't child process, must be parent */
                                                                                    {

                                                                                        /* start of the parent write process; pid_read contains the PID of the child read process */

                                                                                        retval = parent_write (fd, infile); /* parent write loop, process all write records to the modem */

                                                                                        /* end of the parent write process, fall through and exit */

                                                                                    }

                                                                                }

                                                                            }

                                                                        }

                                                                    }

                                                                }

                                                            }

                                                        }

                                                    }

                                                }

                                            }

                                        }

                                    }

                                }

                            }

                        }

                    }

                }

            }

        }

    }

    spclose (0); /* shutdown */
    exit (-retval); /* exit with the error value, formality: spclose () did the exit */

#ifdef LINT

    return (0); /* for lint formalities */

#endif

}

/*

Print any error messages.

static void print_message (int pretval);

I) Data structures:

    A) The help_message array is an array of character stings, one per
       line to be printed for help.

    B) The error_message array is an array of character strings, one
       line per error; the error_message array is implicitly addressed
       by the value integer pretval, which specifies the error to be
       printed.

II) Function execution:

    A) Depending on the value of the integer pretval:

        1) If pretval is zero, print nothing-normal/successful program
           exit.

        2) Else if pretval is unity, print help.

        3) Else pretval is not zero or unity, it is an error code, print
           the corresponding error message.

Returns nothing.

*/

#ifdef __STDC__

static void print_message (int pretval)

#else

static void print_message (pretval)
int pretval;

#endif

{
    size_t help_ctr; /* help_message line counter */

    switch (pretval) /* which return value */
    {

        case 0: /* program ended without errors, print nothing */

            break;

        case 1: /* program ended with a request for help, or argument error, print help */

            for (help_ctr = 0; help_ctr < (sizeof (help_message) / sizeof (char *)); help_ctr ++) /* for each line of help */
            {
                (void) printf ("%s", help_message[help_ctr]); /* print the line */
            }

            break;

        default: /* an error that was not a request for help, print the error */

            (void) fprintf (stderr, "%s", error_message[pretval]);
            break;

    }

}

/*

Parse a record based on sequential delimiters.

int strtoken (char *string, char *parse_array, char **parse, const char *delim);

I) Parse a character array, string, into an array, parse_array, using
   consecutive characters from delim as field delimiters, point the
   character pointers, token, to the beginning of each field.

Returns the number of fields parsed.

*/

#ifdef __STDC__

static int strtoken (char *string, char *parse_array, char **parse, const char *delim)

#else

static int strtoken (string, parse_array, parse, delim)
char *string;
char *parse_array;
char **parse;
char *delim;

#endif

{
    int tokens = 0;

    (void) strcpy (parse_array, string); /* copy the string */

    parse[tokens] = strtok (parse_array, delim); /* get the 1st field */

    while (parse[tokens] != 0) /* get the remaining fields */
    {
        parse[++ tokens] = strtok ((char *) 0, delim);
    }

    return (tokens); /* return the number of tokens parsed */
}

/*

Interrupt handler and shutdown.

static void spclose (int err);

I) Prerequisites:

    A) All file opening and serial port initialization is done by the
       parent write process, (excepting the log file, which opened and
       closed dynamically by the child read process-and is normally
       closed):

        1) The parent write process should close all files, (except
           the log file,) and restore the serial port.

        2) The child read process should close the log file.

    B) Called with SIGINT = ^C, or, SIGTERM, or, SIGCHLD interrupts,
       or, with a negative argument representing an internal error,
       or, an argument of zero for no errors or interrupts.

    C) There are "edge" issues, (i.e., a flag is tested to see if
       something needs shutting down, and the something is shut down,
       but the program is interrupted-perhaps by SIGINT = ^C
       bashing-before the flag gets reset,) There are no error message
       printed in the shut down section of this routine-everything is
       shut down, and at the end of this routine, the flags are
       queried for a successful shutdown, and shutdown errors, if any,
       printed. The idea being that if an "edge" issue occurred, the
       shut down would occur twice, (presumably giving an error the
       second time, which would not be printed, but as long as the
       flag is reset, no error would be printed.)

    D) The kill of the child read by the parent write process, (and
       vice verse,) is complex:

        1) static passed = reset
        2) if parent
            a) if err is SIGCHLD
                i) wait for the child read process to die
            b) if passed = reset
                i) reset all things in the parent process
                ii) print any errors
                iii) set passed
                iv) if err not SIGCHLD and child read process active,
                    kill child read process
        3) else if child
            a) if passed = reset
                i) close the log file
                ii) print any errors
                iii) set passed
                iv) kill parent
        4) restore all signals

II) Data structures:

    A) if the pid is equal to pid_write then the parent write process
       is executing this routine, else, it is the child read process.

    B) The parent write and child read process have a separate copy of
       the int passed flag-which means a complete pass has been made
       through this routine for the client or parent.

III) Function execution:

    A) Upon entry:

        1) If the PID of the process is the parent write process:

            a) Shut down all devices opened by the parent write
               process.

            b) Print any pending errors that occurred before the
               program entered this routine.

            c) Print any errors after the shutdown is complete.

            d) If the call to this routine was not caused by a dead
               child read process, and the child process is actve,
               kill the child read process.

        2) If the PID of the process is the child read process:

            a) Shut down all devices opened by the child read process.

            b) Print any pending errors that occurred before the
               program entered this routine.

            c) Print any errors after the shutdown is complete.

            d) Kill the parent write process.

        3) Restore the signal handlers for SIGINT = ^C and SIGTERM.

There is no return from this routine. The routine exits with a value
of zero if no errors or interrupts occurred, or, the value of the
interrupt, (a positive number,) if an interrupt occurred, or, with a
negative error number if an an error occurred.

*/

#ifdef __STDC__

static void spclose (int err)

#else

static void spclose (err)
int err;

#endif

{
    static int passed = 0; /* passed this routine, 0 = no, 1 = yes, used by both parent write and child read processes, but are different copies */

    int exit_value,
        local_retval = NOERROR; /* assume no error shutting things down */

    if (pid_write == getpid ()) /* parent write process? */
    {

        if (err == SIGCHLD) /* child read process died? */
        {
            (void) wait ((int *) 0); /* free the child read process */
        }

        if (passed == 0) /* passed this routine, 0 = no, 1 = yes? */
        {

            if (state == 1) /* state of serial port setup, 0 = not set up, 1 = original device attributes successfully saved */
            {
                (void) tcflush (fd, TCIOFLUSH); /* flush device input/output */

                if (tcsetattr (fd, TCSANOW, &oldtio) == 0) /* reset the device attributes */
                {
                    state = 0; /* state of serial port setup, 0 = not set up, 1 = original device attributes successfully saved */
                    (void) tcflush (fd, TCIOFLUSH); /* flush device input/output */
                }

            }

            if (fd > 0) /* serial port device file descriptor successfully opened? */
            {
                (void) tcflush (fd, TCIOFLUSH); /* flush device input/output */

                if (close (fd) == 0) /* yes, close the device */
                {
                    fd = -1; /* reset the serial port device file descriptor */
                }

            }

            if (infile != STDIN_FILENO) /* input file open? */
            {

                if (close (infile) == 0) /* close the input file */
                {
                    infile = STDIN_FILENO; /* reset the input file descriptor */
                }

            }

            if (outfile != stdout) /* output file open? */
            {

                if (fclose (outfile) == 0) /* close the output file */
                {
                    outfile = stdout; /* reset the reference to output file */
                }

            }

            if (semaphorelock >= 0) /* lock file open? */
            {
                lockfile.l_type = F_UNLCK; /* unlock the file */
                (void) fcntl (semaphorelock, F_SETLKW, &lockfile); /* unlock the lock file */

                if (close (semaphorelock) == 0) /* close the lock file */
                {
                    semaphorelock = -1; /* reset the lock file file descriptor */

                    if (remove_lockfile == 1) /* remove lock file, 0 = no, 1 = yes, set? */
                    {

                        if (unlink (lockname) == 0) /* remove the lock file */
                        {
                            remove_lockfile = 0; /* remove lock file, 0 = no, 1 = yes */
                        }

                    }

                }

            }

            print_message (-retval); /* print any error messages from the program, (there can only be zero or one) */

            /* print any errors shutting things down */

            if (state != 0) /* state of serial port still setup? */
            {print_message (-ERATTR); local_retval = ERATTR;} /* print the error message, set the local error shutting things down */

            if (semaphorelock != -1) /* lock file open? */
            {print_message (-ELOCKR); local_retval = ELOCKR;} /* print the error message, set the local error shutting things down */

            /* probably don't care about these-exit () will handle them */

            if (fd != -1) /* serial port device file descriptor still open? */
            {print_message (-ECLOSE); local_retval = ECLOSE;} /* print the error message, set the local error shutting things down */

            if (infile != STDIN_FILENO) /* input file open? */
            {print_message (-EICLOSE); local_retval = EICLOSE;} /* print the error message, set the local error shutting things down */

            if (outfile != stdout) /* output file open? */
            {print_message (-EOCLOSE); local_retval = EOCLOSE;} /* print the error message, set the local error shutting things down */

            passed = 1; /* passed this routine, 0 = no, 1 = yes, used by both parent write and child read processes, but are different copies */

            if (err != SIGCHLD) /* child read process died? */
            {

                if (pid_read != (pid_t) 0) /* child read process forked? */
                {
                    (void) kill (pid_read, SIGTERM); /* kill the child read process */
                }

            }

        }

    }

    else /* no, child read process */
    {

        if (passed == 0) /* passed this routine, 0 = no, 1 = yes? */
        {

            if (logfile != (FILE *) 0) /* log file open? */
            {

                if (fclose (logfile) == 0) /* close the log file */
                {
                    logfile = (FILE *) 0; /* reset the reference to log file */
                }

            }

            print_message (-retval); /* print any error messages from the program, (there can only be zero or one) */

            /* probably don't care about this-exit () will handle it */

            if (logfile != (FILE *) 0) /* log file open? */
            {print_message (-ELOGC); local_retval = ELOGC;} /* print the error message, set the local error shutting things down */

            passed = 1; /* passed this routine, 0 = no, 1 = yes, used by both parent write and child read processes, but are different copies */

            if (pid_write != (pid_t) 0) /* parent write process active? */
            {
                (void) kill (pid_write, SIGTERM); /* kill the parent write process */
            }

        }

    }

    /* probably don't care about these-exit () will handle them */

    (void) signal (SIGCHLD, SIG_DFL); /* restore default interrupt for SIGCHLD */
    (void) signal (SIGTERM, SIG_DFL); /* restore default interrupt for SIGTERM */
    (void) signal (SIGINT, SIG_DFL); /* restore default for SIGINT = ^C interrupt */

    exit_value = err; /* assume routine called externally, i.e., SIGINT == ^C or SIGTERM */

    if (exit_value == 0) /* no? */
    {
        exit_value = retval; /* assume error in main body of program, before this routine */

        if (retval == 0) /* no? */
        {
            exit_value = local_retval; /* assume error shutting things down */
        }

    }

    exit (exit_value); /* exit with the value of the interrupt, the value of an error in the main body of this program, the value of an error in this routine, or zero for no error */
}

/*

Parent write loop, process all records from the input file, (usually,
stdin,) to the serial device.

static int parent_write (int fdd, int infiled);

I) Data structure:

    A) The serial port device file descriptor, fdd, should be open, and
       available.

    B) The input file descriptor, infiled, should be open and
       available; it is probably stdin, (but can be a named pipe,
       etc.)

II) Function execution:

    A) This routine is a "spin loop" that waits for a signal, or, an
       error to terminate the program, while reading characters from
       the input file descriptor, infiled:

        1) The number is de-white spaced, and digit data concatinated,
           (i.e., "+1 (987) 654 - 3210" becomes "9876543210")

        2) The phone number in the write buffer is validated to be the
           correct length, and any local area code removed.

        3) The modem dialer prefix and suffix are concatenated to the
           phone number, and written to the serial device file
           descriptor.

Returns any errors encountered, NOERROR on success.

*/

#ifdef __STDC__

static int parent_write (int fdd, int infiled)

#else

static int parent_write (fdd, infiled)
int fdd;
int infiled;

#endif

{
    static char buffer[BUFLEN + 1], /* input write buffer */
                parsebuffer[BUFLEN], /* parsed input write buffer */
                *token[(BUFLEN + 2) / 2], /* reference to tokens in parsed input buffer */
                phone_number[BUFLEN], /* phone number to be dialed */
                record_buffer[BUFLEN]; /* log file record buffer */

    char *number; /* reference to character in phone number */

    int local_retval = NOERROR, /* local return value, (so ERDIN = "Error reading input file" will not go to spclose () on a SIGINT = ^C or SIGTERM "edge" in this routine,) assume no error */
        fields, /* number of fields in the input write buffer */
        field_ctr; /* field counter */

    size_t ac_length = strlen (AREACODE), /* length of local area code */
           length; /* modem dial string length */

    ssize_t cnt; /* count of characters read */

    buffer[BUFLEN] = '\0'; /* write buffer sentinel */

    while (local_retval == NOERROR) /* while no error, spin though getting charcters from the serial device, (actually, waiting for a SIGINT = ^C or SIGTERM) */
    {
        local_retval = ERDIN; /* assume error reading input file */

        if ((cnt = read (infiled, buffer, BUFLEN)) > 0) /* read characters from the input file */
        {
            local_retval = EWROVFL; /* assume error write buffer overflow */

            if ((number = strchr (buffer, '\n')) != (char *) 0) /* buffer have an EOL character before the sentinel? */
            {

                if (((buffer[0] == 'A') || (buffer[0] == 'a')) && ((buffer[1] == 'T') || (buffer[1] == 't'))) /* modem command? */
                {

                    /*

                    UNDOCUMENTED FUNCTIONALITY:

                    A modem command begins with 'AT ...' and the
                    command will be passed directly to the modem-this
                    is useful for commands like 'ATA', (answer a data
                    call,) followed a short time later by 'ATZ0',
                    (reset.)

                    */

                    *number = eol[0]; /* change the EOL character, '\n', to a '\r', a serial port end-of-line string */
                    local_retval = EWRITE; /* assume error writing to serial device */

                    if (write (fdd, buffer, (size_t) cnt) == (ssize_t) cnt) /* write the command to the serial device */
                    {
                        local_retval = EDFLUSH; /* assume error flushing serial device */

                        if (tcdrain (fdd) == 0) /* flush the serial device */
                        {
                            local_retval = NOERROR; /* assume no error */
                        }

                    }

                }

                else
                {
                    *number = '\0'; /* remove the EOL character */
                    local_retval = NOERROR; /* assume no error */

                    if ((fields = strtoken (buffer, parsebuffer, token, NUMBER_TOKEN_SEPARATORS)) > 0) /* parse the record into fields, skip the record if there are no fields */
                    {
                        phone_number[0] = '\0'; /* null the phone number to be dialed */

                        for (field_ctr = 0; field_ctr < fields; field_ctr ++) /* for each field in the input record */
                        {
                            (void) strcat (phone_number, token[field_ctr]); /* concatinate the fields, i.e., "(987) 654-3210" would become "9876543210" */
                        }

                        number = phone_number; /* reference the beginning of the phone number */

                        do /* number have anthing but digits? */
                        {

                            if (isdigit (*number) == 0) /* character a digit */
                            {
                                local_retval = EPDIAL; /* error processing dial record */
                                break;
                            }

                        }
                        while (*(++ number) != '\0'); /* next character */

                        if (local_retval == NOERROR) /* any errors */
                        {
                            number = phone_number; /* reference the beginning of the phone number */
                            local_retval = EPDIAL; /* assume error processing dial record */

                            switch (strlen (number)) /* check for valid phone number */
                            {

                                case 0: /* too few digits for a phone number */
                                case 1:
                                case 2:
                                case 3:
                                case 4:
                                case 5:
                                case 6:

                                    break;

                                case 7: /* possibly, a local number */

                                    local_retval = NOERROR; /* assume no error */
                                    break;

                                case 8: /* wrong number of digits for a phone number */
                                case 9:

                                    break;

                                case 10: /* possibly, a long distance number, without the leading '1' */

                                    if (strncmp (number, AREACODE, ac_length) == 0) /* long distance number start with the local area code? */
                                    {
                                        number = number + ac_length; /* skip over the area code, the local number begins after the local area code */
                                    }

                                    else /* possibly, a long distance number, without the leading '1' */
                                    {
                                        memmove (&number[1], number, 10); /* move the number right, to make a place for the leading '1' */
                                        *number = '1'; /* include the '1' */
                                    }

                                    local_retval = NOERROR; /* assume no error */
                                    break;

                                case 11: /* possibly, a long distance number, with the leading '1' */

                                    if (number[0] == '1') /* all 11 digit numbers must begin with '1', long distance number begin with a '1'? */
                                    {

                                        if (strncmp (&number[1], AREACODE, ac_length) == 0) /* long distance number start with a '1' plus the local area code? */
                                        {
                                            number = number + ac_length + 1; /* skip over the area code, the local number begins after the '1' plus the local area code */
                                        }

                                        local_retval = NOERROR; /* assume no error */
                                    }

                                    break;

                                default: /* possibly a foreign number, with country code */

                                    local_retval = NOERROR; /* assume no error */
                                    break;
                            }

                            if (local_retval == NOERROR) /* valid phone number? */
                            {

                                if (debug == 1) /* debug flag, 0 = no, 1 = yes set? */
                                {
                                    (void) sprintf (record_buffer, "Write debug: \"%s%s%s\"\n", dialprefix, number, dialsuffix); /* concatenate the output file record */
                                    local_retval = EOWRITE; /* assume error writing to output file */

                                    if (fputs (record_buffer, outfile) > 0) /* print the record to the output file */
                                    {
                                        local_retval = EOFLUSH; /* assume error flushing output file */

                                        if (fflush (outfile) == 0) /* flush the output file */
                                        {
                                            local_retval = NOERROR; /* assume no error */
                                        }

                                    }

                                }

                                if (local_retval == NOERROR) /* any errors? */
                                {
                                    local_retval = EWRITE; /* assume error writing to serial device */
                                    (void) sprintf (record_buffer, "%s%s%s%s", dialprefix, number, dialsuffix, eol); /* concatenate the modem dial string */
                                    length = strlen (record_buffer); /* save the modem dial string length */

                                    if (write (fdd, record_buffer, length) == (ssize_t) length) /* write the number to the serial device */
                                    {
                                        local_retval = EDFLUSH; /* assume error flushing serial device */

                                        if (tcdrain (fdd) == 0) /* flush the serial device */
                                        {
                                            local_retval = NOERROR; /* assume no error */
                                        }

                                    }

                                }

                            }

                        }

                    }

                }

            }

        }

    }

    return (retval = local_retval); /* return any errors */
}

/*

Child read loop, process all records from the modem.

static int child_read (int fdd);

I) Data structure:

    A) The serial port device file descriptor, fdd, should be open, and
       available.

    B) The CID record fields from the modem:

        1) "DATE = ", MMDD
        2) "TIME = ", HHMM, 24 hour
        3) "NAME = ", variable length
        4) "NMBR = ", 10 digit phone number, (no leading 1)

    C) All other records from the modem reset all CID data.

II) Function execution:

    A) This routine is a "spin loop" that waits for a signal, or, an
       error to terminate the program, while reading characters from
       the serial device file descriptor, fdd:

        1) The first token is searched for a match of CID field, and
           if found, the data saved, and the read_state is incremented
           by multiple of ten, depending on the field.

        2) When the read_state totals 1111, a complete complement of
           CID data has been found, and the data formatted, and output
           to the output file, (usually stdout,) and the log file.

Returns any errors encountered, NOERROR on success.

*/

#ifdef __STDC__

static int child_read (int fdd)

#else

static int child_read (fdd)
int fdd;

#endif

{
    static char buffer[BUFLEN + 1], /* input read buffer */
                parsebuffer[BUFLEN], /* parsed input read buffer */
                *token[(BUFLEN + 2) / 2], /* reference to tokens in parsed input buffer */
                date_buffer[BUFLEN], /* buffer for CID date */
                time_buffer[BUFLEN], /* buffer for CID time */
                name_buffer[BUFLEN], /* buffer for CID name */
                number_buffer[BUFLEN], /* buffer for CID number */
                record_buffer[BUFLEN]; /* log file record buffer */

    char *number; /* reference to character in phone number */

    int local_retval = NOERROR, /* local return value, (so EREAD = "Error reading from device" will not go to spclose () on a SIGINT = ^C or SIGTERM "edge" in this routine,) assume no error */
        read_state = 0, /* state counter; 0 = initial state, 1 = DATE found, 10 = TIME found, 100 = NAME found, 1000 = NUMBER found, OR'ed together, i.e., state = 1111 means a complete compliment of a CID */
        fields, /* number of fields in the input read buffer */
        field; /* field number */

    ssize_t cnt; /* count of characters read */

    time_t unix_time; /* time stamp structure */

    buffer[BUFLEN] = '\0'; /* write buffer sentinel */

    while (local_retval == NOERROR) /* while no error, spin though getting charcters from the serial device, (actually, waiting for a SIGINT = ^C or SIGTERM) */
    {
        local_retval = EREAD; /* assume error reading from device */

        if ((cnt = read (fdd, buffer, BUFLEN)) > 0) /* read characters from the serial device */
        {
            local_retval = EREOVFL; /* assume error read buffer overflow */

            if ((number = strchr (buffer, '\n')) != (char *) 0) /* buffer have an EOL character before the sentinel? */
            {
                *number = '\0'; /* remove the EOL character */
                local_retval = NOERROR; /* assume no error */

                if ((fields = strtoken (buffer, parsebuffer, token, TOKEN_SEPARATORS)) > 0) /* parse the record into fields, skip the record if there are no fields */
                {

                    if (strcmp (token[0], "DATE") == 0) /* CID date? */
                    {
                        (void) strcpy (date_buffer, token[2]); /* yes, save the date string */
                        read_state = read_state + 1; /* state counter; 0 = initial state, 1 = DATE found, 10 = TIME found, 100 = NAME found, 1000 = NUMBER found, OR'ed together, i.e., state = 1111 means a complete compliment of a CID */
                    }

                    else if (strcmp (token[0], "TIME") == 0) /* CID time? */
                    {
                        (void) strcpy (time_buffer, token[2]); /* yes, save the time string */
                        read_state = read_state + 10; /* state counter; 0 = initial state, 1 = DATE found, 10 = TIME found, 100 = NAME found, 1000 = NUMBER found, OR'ed together, i.e., state = 1111 means a complete compliment of a CID */
                    }

                    else if (strcmp (token[0], "NAME") == 0) /* CID name? */
                    {
                        name_buffer[0] = '\0'; /* null the CID name */

                        for (field = 2; field < fields; field ++) /* for each field in the CID name */
                        {
                            (void) strcat (name_buffer, token[field]); /* add the field in the CID name */

                            if (field < (fields - 1)) /* another field in the CID name? */
                            {
                                (void) strcat (name_buffer, " "); /* add a space word delimiter */
                            }

                        }

                        read_state = read_state + 100; /* state counter; 0 = initial state, 1 = DATE found, 10 = TIME found, 100 = NAME found, 1000 = NUMBER found, OR'ed together, i.e., state = 1111 means a complete compliment of a CID */
                    }

                    else if (strcmp (token[0], "NMBR") == 0) /* CID number */
                    {
                        (void) strcpy (number_buffer, token[2]); /* yes, save the number string */
                        read_state = read_state + 1000; /* state counter; 0 = initial state, 1 = DATE found, 10 = TIME found, 100 = NAME found, 1000 = NUMBER found, OR'ed together, i.e., state = 1111 means a complete compliment of a CID */
                    }

                    else if (strcmp (token[0], "RING") == 0) /* RING response from modem? */
                    {
                        read_state = 0; /* state counter; 0 = initial state, 1 = DATE found, 10 = TIME found, 100 = NAME found, 1000 = NUMBER found, OR'ed together, i.e., state = 1111 means a complete compliment of a CID */
                    }

                    if (debug == 1) /* debug flag, 0 = no, 1 = yes set? */
                    {
                        (void) sprintf (record_buffer, "Read debug: %s\n", buffer); /* concatenate the output file record */
                        local_retval = EOWRITE; /* assume error writing to output file */

                        if (fputs (record_buffer, outfile) > 0) /* print the record to the output file */
                        {

                            if (fflush (outfile) == 0) /* flush the output file */
                            {
                                local_retval = NOERROR; /* assume no error */
                            }

                        }

                    }

                    if (local_retval == NOERROR) /* any errors? */
                    {

                        if (read_state == 1111) /* complete compliment of a CID? */
                        {
                            local_retval = EOWRITE; /* assume error writing to output file */
                            (void) sprintf (record_buffer, "--Call Logged--\n"); /* concatenate the output file record */

                            if (fputs (record_buffer, outfile) > 0) /* print the record to the output file */
                            {
                                (void) sprintf (record_buffer, "  Date:   %s\n", date_buffer); /* concatenate the output file record */

                                if (fputs (record_buffer, outfile) > 0) /* print the record to the output file */
                                {
                                    (void) sprintf (record_buffer, "  Time:   %s\n", time_buffer); /* concatenate the output file record */

                                    if (fputs (record_buffer, outfile) > 0) /* print the record to the output file */
                                    {
                                        (void) sprintf (record_buffer, "  Name:   %s\n", name_buffer); /* concatenate the output file record */

                                        if (fputs (record_buffer, outfile) > 0) /* print the record to the output file */
                                        {
                                            (void) sprintf (record_buffer, "  Number: %s\n", number_buffer); /* concatenate the output file record */

                                            if (fputs (record_buffer, outfile) > 0) /* print the record to the output file */
                                            {
                                                local_retval = EOFLUSH; /* assume error flushing output file */

                                                if (fflush (outfile) == 0) /* flush the output file */
                                                {
                                                    local_retval = EEPOC; /* assume error getting Epoc time */

                                                    if ((unix_time = time ((time_t *) 0)) > (time_t) 0) /* get the time stamp */
                                                    {
                                                        (void) sprintf (record_buffer, "%s\t%s\t%s\t%s\t%d\n", date_buffer, time_buffer, name_buffer, number_buffer, (int) unix_time); /* concatenate the log file record */
                                                        local_retval = ELOGO; /* assume error opening log file */

                                                        if ((logfile = fopen (logname, "a")) != (FILE *) 0) /* open the log file */
                                                        {
                                                            local_retval = ELOGW; /* assume error writing to log file */

                                                            if (fputs (record_buffer, logfile) > 0) /* print the record to the log file */
                                                            {
                                                                local_retval = ELOGC; /* assume error closing log file */

                                                                if (fclose (logfile) == 0)
                                                                {
                                                                    logfile = (FILE *) 0; /* reset the reference to log file */
                                                                    local_retval = NOERROR; /* assume no error */
                                                                }

                                                            }

                                                        }

                                                    }

                                                }

                                            }

                                        }

                                    }

                                }

                            }

                            read_state = 0; /* state counter; 0 = initial state, 1 = DATE found, 10 = TIME found, 100 = NAME found, 1000 = NUMBER found, OR'ed together, i.e., state = 1111 means a complete compliment of a CID */
                        }

                    }

                }

            }

        }

    }

    return (retval = local_retval); /* return any errors */
}
